(*====================================================================
  =  Program Id:     CROSS.MOD                                       =
  =  Programmer:     Peter M. Perchansky                             =
  =  Purpose:        Source Code Lister with Cross Reference.        =
  =  Date Written:   12-23-89                                        =
  =------------------------------------------------------------------=
  =                       Version History                            =
  =------------------------------------------------------------------=
  =  01-12-90  PMP  Expanded code set to handle 177C to 377C.        =
  =                 Thank you Scott McIntosh.                        =
  =  12-25-89  PMP  Cleaned up code.                                 =
  =  12-23-89  PMP  Date written.                                    =
  ====================================================================*)

(*$M CODE *)
MODULE Cross;

(*------------- Procedures from JPI's TopSpeed Modula II -------------*)
              IMPORT IO;
FROM FIO      IMPORT Close, Create, EOF, Exists, File, Open;
FROM FIO      IMPORT RdChar, WrChar, WrCard, WrStr, WrLn;
FROM Lib      IMPORT ParamCount, ParamStr;
FROM Storage  IMPORT ALLOCATE, Available;
FROM Str      IMPORT Append, Copy, Pos, Slice;
(*--------------------------------------------------------------------*)

(*------------- Procedures from PMP's Library Modules ----------------*)
FROM PMPAsm   IMPORT Cls, GotoXY;
FROM PMPError IMPORT FatalError;
(*--------------------------------------------------------------------*)


CONST
    bufferLength = 10000;
    maxCard      = MAX (CARDINAL);
    wordLength   = 16;

TYPE
    itemPtr = POINTER TO item;
    wordPtr = POINTER TO word;

    word    = RECORD
                key    : CARDINAL;
                first,
                last   : itemPtr;
                left,
                right  : wordPtr;
              END;

    item    = RECORD
                lno    : CARDINAL;
                next   : itemPtr;
              END;

VAR
    root       : wordPtr;
    k0, k1,
    line       : CARDINAL;
    character  : CHAR;
    buffer     : ARRAY [0..bufferLength - 1] OF CHAR;
    inFile,
    outFile    : File;

PROCEDURE PrintWord (k: CARDINAL);

VAR
    limit : CARDINAL;

BEGIN
    limit := k + wordLength;
    WHILE buffer[k] > 0C DO
          WrChar (outFile, buffer[k]);
          INC (k);
    END;

    WHILE k < limit DO
          WrChar (outFile, ' ');
          INC (k);
    END;
END PrintWord;

PROCEDURE PrintTree (t: wordPtr);

VAR
    i, m  : INTEGER;
    item  : itemPtr;

BEGIN
    IF t # NIL THEN
       WITH t^ DO
            PrintTree (left);
            PrintWord (key);
            item := first;
            m := 0;
            REPEAT
                 IF m = 8 THEN
                    WrLn (outFile);
                    m := 0;

                    FOR i := 1 TO wordLength DO
                        WrChar (outFile, ' ')
                    END;
                 END;

                 INC (m);
                 WrCard (outFile, item^.lno, 5);
                 item := item^.next;
            UNTIL item = NIL;

            WrLn (outFile);
            PrintTree (right);
       END;
    END;
END PrintTree;

PROCEDURE Diff (i, j: CARDINAL): INTEGER;

BEGIN
    LOOP
       IF buffer[i] # buffer[j] THEN
          RETURN (INTEGER (ORD (buffer[i])) - INTEGER (ORD (buffer[j])))
       ELSIF buffer[i] = 0C THEN
             RETURN 0
          END;

          INC (i);
          INC (j);
    END;
END Diff;

PROCEDURE Search (VAR p: wordPtr);

VAR
    d     : INTEGER;
    item  : itemPtr;

BEGIN
    IF p = NIL THEN (* word not in tree; insert *)
       IF NOT Available (SIZE (word) + SIZE (item)) THEN
          FatalError ("Not enough memory to allocate new items!")
       ELSE
          ALLOCATE (p, SIZE (word));
          ALLOCATE (item, SIZE (item));
          WITH p^ DO
               key := k0;
               first := item;
               last := item;
               left := NIL;
               right := NIL;
          END;

          item^.lno := line;
          item^.next := NIL;
          k0 := k1;
       END;
    ELSE
       d := Diff (k0, p^.key);
       IF d < 0 THEN
          Search (p^.left)
       ELSIF d > 0 THEN
             Search (p^.right)
          ELSIF NOT Available (SIZE (word) + SIZE (item)) THEN
                FatalError ("Not enough memory to allocate new items!")
             ELSE
                ALLOCATE (item, SIZE (item));
                item^.lno := line;
                item^.next := NIL;
                p^.last^.next := item;
                p^.last := item;
             END;
    END;
END Search;

PROCEDURE GetWord ();

BEGIN
    k1 := k0;
    REPEAT
         WrChar (outFile, character);
         buffer[k1] := character;
         INC (k1);
         character := RdChar (inFile);
    UNTIL (character < "0") OR (character > "9")
      AND (CAP (character) < 'A') OR (CAP (character) > 'Z');

    buffer[k1] := 0C;
    INC (k1); (* terminator *)
    Search (root);
END GetWord;

PROCEDURE GetFiles (VAR in, out: File);

VAR
    input    : ARRAY [0..79] OF CHAR;
    location : CARDINAL;

BEGIN
    Cls ();

    IF ParamCount () = 0 THEN
       REPEAT
            IO.WrLn;
            IO.WrStr (" Please enter input file:  ");
            IO.RdStr (input);
       UNTIL Exists (input);
    ELSIF ParamCount () > 0 THEN
          ParamStr (input, 1);
          location := Pos (input, '.');

          IF location = maxCard THEN
             Append (input, ".MOD")
          END;

          WHILE NOT Exists (input) DO
                IO.WrLn;
                IO.WrStr (" Please enter input file:  ");
                IO.RdStr (input);
          END;
       END;

    in := Open (input);

    IF ParamCount () = 2 THEN
       ParamStr (input, 2);
       location := Pos (input, '.');
       IF location = maxCard THEN
          Append (input, ".XRF")
       END;
    ELSE
       location := Pos (input, '.');
       IF location = maxCard THEN
          Append (input, ".XRF")
       ELSE
          Slice (input, input, 0, location);
          Append (input, ".XRF");
       END;
    END;

    out := Create (input);
END GetFiles;

BEGIN
    root := NIL;
    k0 := 0;
    line := 0;

    GetFiles (inFile, outFile);

    Cls ();
    GotoXY (10, 25);
    IO.WrStr ("Processing...");

    WrCard (outFile, 0, 5);
    WrChar (outFile, ' ');

    character := RdChar (inFile);
    WHILE NOT EOF DO
          CASE character OF
            |  0C..11C   : character := RdChar (inFile)
            |      12C   : WrLn (outFile);  (* line feed *)
                           character := RdChar (inFile);

                           IF NOT EOF THEN
                              INC (line);
                              WrCard (outFile, line, 5);
                              WrChar (outFile, ' ');
                           END;
            |  13C..37C  : character := RdChar (inFile)
            |  40C..100C : WrChar (outFile, character);
                           character := RdChar (inFile);
            | 101C..132C : GetWord ()
            | 133C..140C : WrChar (outFile, character);
                           character := RdChar (inFile);
            | 141C..172C : GetWord ()
            | 173C..377C : WrChar (outFile, character);
                           character := RdChar (inFile);
          END;
    END;

    WrLn (outFile);
    WrLn (outFile);
    Close (inFile);
    PrintTree (root);
    Close (outFile);
END Cross.
